@@ -1,11 +1,21 @@
+/**
+ * gam_error.c: the debugging intrastructure
+ *
+ * Allow to use GAM_DEBUG environment variable or SIG_USR2 for
+ * dynamic debugging of clients and server.
+ *
+ * Daniel Veillard <veillard@redhat.com>
+ * See the Copyright file.
+ */
 #include <config.h>
 #include <stdio.h>
 #include <stdarg.h>
 #include <stdlib.h>
 #include <signal.h>
+#include <errno.h>
 #include "gam_error.h"
 
-typedef void (*signal_handler)(int);
+typedef void (*signal_handler) (int);
 
 extern void gam_show_debug(void);
 
@@ -16,36 +26,45 @@ static int got_signal = 0;
 static FILE *debug_out = NULL;
 
 static void
-gam_error_handle_signal(void) {
+gam_error_handle_signal(void)
+{
     if (got_signal == 0)
         return;
+
     got_signal = 0;
 
     if (do_debug == 0) {
-       char path[50] = "/tmp/gamin_debug_XXXXXX";
-       int fd = mkstemp(path);
-       if (fd >= 0) {
-	   debug_out = fdopen(fd, "a");
-	   if (debug_out != NULL) {
-	       do_debug = 1;
-	       gam_debug_active = 1;
-	       gam_show_debug();
-	   }
-       }
+        if (debug_out != stderr) {
+            char path[50] = "/tmp/gamin_debug_XXXXXX";
+            int fd = mkstemp(path);
+
+            if (fd >= 0) {
+                debug_out = fdopen(fd, "a");
+                if (debug_out != NULL) {
+                    do_debug = 1;
+                    gam_debug_active = 1;
+                    gam_show_debug();
+                }
+            }
+        }
     } else {
-       do_debug = 0;
-       gam_debug_active = 0;
-       if (debug_out != NULL) {
-           fclose(debug_out);
-	   debug_out = NULL;
-       }
+        if (debug_out != stderr) {
+            do_debug = 0;
+            gam_debug_active = 0;
+            if (debug_out != NULL) {
+                fflush(debug_out);
+                fclose(debug_out);
+                debug_out = NULL;
+            }
+        }
     }
 }
 
-static void 
-gam_error_signal(int no) {
+static void
+gam_error_signal(int no)
+{
     got_signal = !got_signal;
-    gam_debug_active = -1; /* force going into gam_debug() */
+    gam_debug_active = -1;      /* force going into gam_debug() */
 }
 
 /**
@@ -54,21 +73,27 @@ gam_error_signal(int no) {
  * Initialization routine for the error and debug handling.
  */
 void
-gam_error_init(void) {
-
+gam_error_init(void)
+{
     if (initialized == 0) {
-	signal_handler prev;
+        signal_handler prev;
 
         initialized = 1;
+
         if (getenv("GAM_DEBUG") != NULL) {
-            do_debug = 1;
+	    debug_out = stderr;
 	    gam_debug_active = 1;
-	}
-	prev = signal(SIGUSR2, gam_error_signal);
-	/* if there is already an handler switch back to the original
-	   to avoid disturbing the application behaviour */
-	if ((prev != SIG_IGN) && (prev != SIG_DFL) && (prev != NULL))
-	    signal(SIGUSR2, prev);
+	    do_debug = 1;
+            /* Fake the signal */
+            got_signal = 1;
+            gam_error_handle_signal();
+        }
+
+        prev = signal(SIGUSR2, gam_error_signal);
+        /* if there is already an handler switch back to the original
+         * to avoid disturbing the application behaviour */
+        if ((prev != SIG_IGN) && (prev != SIG_DFL) && (prev != NULL))
+            signal(SIGUSR2, prev);
     }
 }
 
@@ -79,13 +104,19 @@ gam_error_init(void) {
  * error debugging events.
  */
 void
-gam_error_check(void) {
+gam_error_check(void)
+{
     if (initialized == 0)
         gam_error_init();
 
     if (got_signal)
         gam_error_handle_signal();
+}
 
+int
+gam_errno(void)
+{
+    return (errno);
 }
 
 /**
@@ -112,11 +143,13 @@ gam_error(const char *file, int line, const char *function,
 
     if ((file == NULL) || (function == NULL) || (format == NULL))
         return;
+
     va_start(args, format);
     vfprintf((debug_out ? debug_out : stderr), format, args);
     va_end(args);
+
     if (debug_out)
-	fflush(debug_out);
+        fflush(debug_out);
 }
 
 /**
@@ -143,11 +176,13 @@ gam_debug(const char *file, int line, const char *function,
 
     if ((do_debug == 0) || (gam_debug_active == 0))
         return;
+
     if ((file == NULL) || (function == NULL) || (format == NULL))
         return;
+
     va_start(args, format);
     vfprintf((debug_out ? debug_out : stdout), format, args);
     va_end(args);
     if (debug_out)
-	fflush(debug_out);
+        fflush(debug_out);
 }
