@@ -13,6 +13,7 @@
 #include <sys/stat.h>
 #include <sys/socket.h>
 #include <sys/un.h>
+#include <sys/uio.h>
 #include "fam.h"
 #include "gam_protocol.h"
 #include "gam_data.h"
@@ -29,7 +30,7 @@ void *debug_userData = NULL;
 #endif
 int FAMErrno = 0;
 
-enum {
+static enum {
     FAM_OK = 0,
     FAM_ARG,	/* Bad arguments */
     FAM_FILE,	/* Bad filename */
@@ -168,7 +169,7 @@ gamin_get_socket_path(void)
 static char *
 gamin_get_socket_dir(void)
 {
-    char *user;
+    const char *user;
     char *ret;
     char path[MAXPATHLEN + 1];
 
@@ -181,7 +182,6 @@ gamin_get_socket_dir(void)
     snprintf(path, MAXPATHLEN, "/tmp/fam-%s", user);
     path[MAXPATHLEN] = 0;
     ret = strdup(path);
-    free(user);
     return (ret);
 }
 
@@ -421,9 +421,35 @@ gamin_write_credential_byte(int fd)
 {
     char data[2] = { 0, 0 };
     int written;
+#if defined(HAVE_CMSGCRED) && !defined(LOCAL_CREDS)
+    struct {
+	    struct cmsghdr hdr;
+	    struct cmsgcred cred;
+    } cmsg;
+    struct iovec iov;
+    struct msghdr msg;
+
+    iov.iov_base = &data[0];
+    iov.iov_len = 1;
+
+    memset (&msg, 0, sizeof (msg));
+    msg.msg_iov = &iov;
+    msg.msg_iovlen = 1;
+
+    msg.msg_control = &cmsg;
+    msg.msg_controllen = sizeof (cmsg);
+    memset (&cmsg, 0, sizeof (cmsg));
+    cmsg.hdr.cmsg_len = sizeof (cmsg);
+    cmsg.hdr.cmsg_level = SOL_SOCKET;
+    cmsg.hdr.cmsg_type = SCM_CREDS;
+#endif
 
 retry:
+#if defined(HAVE_CMSGCRED) && !defined(LOCAL_CREDS)
+    written = sendmsg(fd, &msg, 0);
+#else
     written = write(fd, &data[0], 1);
+#endif
     if (written < 0) {
         if (errno == EINTR)
             goto retry;
@@ -494,20 +520,25 @@ static int
 gamin_write_byte(int fd, const char *data, size_t len)
 {
     int written;
+    int remaining;
+
+    remaining = len;
+    do {
+	written = write(fd, data, remaining);
+	if (written < 0) {
+	    if (errno == EINTR)
+		continue;
+
+	    GAM_DEBUG(DEBUG_INFO,
+		      "%s: Failed to write bytes to socket %d: %s\n",
+		      __FUNCTION__, fd, strerror (errno));
+	    return -1;
+	}
+
+	data += written;
+	remaining -= written;
+    } while (remaining > 0);
 
-  retry:
-    written = write(fd, &data[0], len);
-    if (written < 0) {
-        if (errno == EINTR)
-            goto retry;
-        gam_error(DEBUG_INFO, "Failed to write bytes to socket %d\n", fd);
-        return (-1);
-    }
-    if (written != (int) len) {
-        gam_error(DEBUG_INFO, "Wrote only %d bytes to socket %d\n",
-                  written, fd);
-        return (-1);
-    }
     GAM_DEBUG(DEBUG_INFO, "Wrote %d bytes to socket %d\n", written, fd);
     return (0);
 }
@@ -616,8 +647,10 @@ gamin_check_cred(GAMDataPtr conn, int fd)
     gid_t c_gid;
 
 #ifdef HAVE_CMSGCRED
-    char cmsgmem[CMSG_SPACE(sizeof(struct cmsgcred))];
-    struct cmsghdr *cmsg = (struct cmsghdr *) cmsgmem;
+    struct {
+	    struct cmsghdr hdr;
+	    struct cmsgcred cred;
+    } cmsg;
 #endif
 
     s_uid = getuid();
@@ -642,9 +675,9 @@ gamin_check_cred(GAMDataPtr conn, int fd)
     msg.msg_iovlen = 1;
 
 #ifdef HAVE_CMSGCRED
-    memset(cmsgmem, 0, sizeof(cmsgmem));
-    msg.msg_control = cmsgmem;
-    msg.msg_controllen = sizeof(cmsgmem);
+    memset(&cmsg, 0, sizeof(cmsg));
+    msg.msg_control = &cmsg;
+    msg.msg_controllen = sizeof(cmsg);
 #endif
 
 retry:
@@ -661,7 +694,7 @@ retry:
         goto failed;
     }
 #ifdef HAVE_CMSGCRED
-    if (cmsg->cmsg_len < sizeof(cmsgmem) || cmsg->cmsg_type != SCM_CREDS) {
+    if (cmsg.hdr.cmsg_len < sizeof(cmsg) || cmsg.hdr.cmsg_type != SCM_CREDS) {
         GAM_DEBUG(DEBUG_INFO,
                   "Message from recvmsg() was not SCM_CREDS\n");
         goto failed;
@@ -673,7 +706,7 @@ retry:
     {
 #ifdef SO_PEERCRED
         struct ucred cr;
-        int cr_len = sizeof(cr);
+        socklen_t cr_len = sizeof(cr);
 
         if (getsockopt(fd, SOL_SOCKET, SO_PEERCRED, &cr, &cr_len) ==
             0 && cr_len == sizeof(cr)) {
@@ -687,13 +720,9 @@ retry:
             goto failed;
         }
 #elif defined(HAVE_CMSGCRED)
-        struct cmsgcred *cred;
-
-        cred = (struct cmsgcred *) CMSG_DATA(cmsg);
-
-        c_pid = cred->cmcred_pid;
-        c_uid = cred->cmcred_euid;
-        c_gid = cred->cmcred_groups[0];
+        c_pid = cmsg.cred.cmcred_pid;
+        c_uid = cmsg.cred.cmcred_euid;
+        c_gid = cmsg.cred.cmcred_groups[0];
 #else /* !SO_PEERCRED && !HAVE_CMSGCRED */
         GAM_DEBUG(DEBUG_INFO,
                   "Socket credentials not supported on this OS\n");
@@ -1012,6 +1041,8 @@ FAMClose(FAMConnection * fc)
 
     GAM_DEBUG(DEBUG_INFO, "FAMClose()\n");
 
+    gamin_data_lock(fc->client);
+
     ret = close(fc->fd);
     fc->fd = -1;
     gamin_data_free(fc->client);
@@ -1033,6 +1064,8 @@ int
 FAMMonitorDirectory(FAMConnection * fc, const char *filename,
                     FAMRequest * fr, void *userData)
 {
+    int retval;
+     
     if ((fc == NULL) || (filename == NULL) || (fr == NULL)) {
 	GAM_DEBUG(DEBUG_INFO, "FAMMonitorDirectory() arg error\n");
         FAMErrno = FAM_ARG;
@@ -1049,8 +1082,13 @@ FAMMonitorDirectory(FAMConnection * fc, const char *filename,
         FAMErrno = FAM_ARG;
         return (-1);
     }
-    return (gamin_send_request(GAM_REQ_DIR, fc->fd, filename,
+    
+    gamin_data_lock(fc->client);
+    retval = (gamin_send_request(GAM_REQ_DIR, fc->fd, filename,
                                fr, userData, fc->client, 0));
+    gamin_data_unlock(fc->client);
+
+    return retval;
 }
 
 /**
@@ -1067,6 +1105,8 @@ int
 FAMMonitorDirectory2(FAMConnection * fc, const char *filename,
                      FAMRequest * fr)
 {
+    int retval;
+
     if ((fc == NULL) || (filename == NULL) || (fr == NULL)) {
 	GAM_DEBUG(DEBUG_INFO, "FAMMonitorDirectory2() arg error\n");
         FAMErrno = FAM_ARG;
@@ -1085,8 +1125,12 @@ FAMMonitorDirectory2(FAMConnection * fc, const char *filename,
         return (-1);
     }
 
-    return (gamin_send_request(GAM_REQ_DIR, fc->fd, filename,
+    gamin_data_lock(fc->client);
+    retval = (gamin_send_request(GAM_REQ_DIR, fc->fd, filename,
                                fr, NULL, fc->client, 1));
+    gamin_data_unlock(fc->client);
+
+    return retval;
 }
 
 /**
@@ -1104,6 +1148,8 @@ int
 FAMMonitorFile(FAMConnection * fc, const char *filename,
                FAMRequest * fr, void *userData)
 {
+    int retval;
+
     if ((fc == NULL) || (filename == NULL) || (fr == NULL)) {
 	GAM_DEBUG(DEBUG_INFO, "FAMMonitorFile() arg error\n");
         FAMErrno = FAM_ARG;
@@ -1120,8 +1166,13 @@ FAMMonitorFile(FAMConnection * fc, const char *filename,
         FAMErrno = FAM_ARG;
         return (-1);
     }
-    return (gamin_send_request(GAM_REQ_FILE, fc->fd, filename,
+
+    gamin_data_lock(fc->client);
+    retval =  (gamin_send_request(GAM_REQ_FILE, fc->fd, filename,
                                fr, userData, fc->client, 0));
+    gamin_data_unlock(fc->client);
+
+    return retval;
 }
 
 /**
@@ -1137,6 +1188,8 @@ FAMMonitorFile(FAMConnection * fc, const char *filename,
 int
 FAMMonitorFile2(FAMConnection * fc, const char *filename, FAMRequest * fr)
 {
+    int retval;
+
     if ((fc == NULL) || (filename == NULL) || (fr == NULL)) {
 	GAM_DEBUG(DEBUG_INFO, "FAMMonitorFile2() arg error\n");
         FAMErrno = FAM_ARG;
@@ -1153,8 +1206,13 @@ FAMMonitorFile2(FAMConnection * fc, const char *filename, FAMRequest * fr)
         FAMErrno = FAM_ARG;
         return (-1);
     }
-    return (gamin_send_request(GAM_REQ_FILE, fc->fd, filename,
+
+    gamin_data_lock(fc->client);
+    retval = (gamin_send_request(GAM_REQ_FILE, fc->fd, filename,
                                fr, NULL, fc->client, 1));
+    gamin_data_unlock(fc->client);
+
+    return retval;
 }
 
 /**
@@ -1223,6 +1281,8 @@ FAMNextEvent(FAMConnection * fc, FAMEvent * fe)
         return (-1);
     }
 
+    // FIXME: drop and reacquire lock while blocked?
+    gamin_data_lock(conn);
     if (!gamin_data_event_ready(conn)) {
         if (gamin_read_data(conn, fc->fd, 1) < 0) {
 	    gamin_try_reconnect(conn, fc->fd);
@@ -1231,6 +1291,8 @@ FAMNextEvent(FAMConnection * fc, FAMEvent * fe)
 	}
     }
     ret = gamin_data_read_event(conn, fe);
+    gamin_data_unlock(conn);
+
     if (ret < 0) {
         FAMErrno = FAM_CONNECT;
         return (ret);
@@ -1270,8 +1332,11 @@ FAMPending(FAMConnection * fc)
 
     GAM_DEBUG(DEBUG_INFO, "FAMPending(fd = %d)\n", fc->fd);
 
-    if (gamin_data_event_ready(conn))
-      return (1);
+    gamin_data_lock(conn);
+    if (gamin_data_event_ready(conn)) {
+	 gamin_data_unlock(conn);
+	 return (1);
+    }
 
     /*
      * make sure we won't block if reading
@@ -1285,7 +1350,10 @@ FAMPending(FAMConnection * fc)
 	}
     }
 
-    return (gamin_data_event_ready(conn));
+    ret = (gamin_data_event_ready(conn));
+    gamin_data_unlock(conn);
+
+    return ret;
 }
 
 /**
@@ -1300,8 +1368,8 @@ FAMPending(FAMConnection * fc)
 int
 FAMCancelMonitor(FAMConnection * fc, const FAMRequest * fr)
 {
-    GAMDataPtr conn;
     int ret;
+    GAMDataPtr conn;
 
     if ((fc == NULL) || (fr == NULL)) {
 	GAM_DEBUG(DEBUG_INFO, "FAMCancelMonitor() arg error\n");
@@ -1317,12 +1385,14 @@ FAMCancelMonitor(FAMConnection * fc, const FAMRequest * fr)
     GAM_DEBUG(DEBUG_INFO, "FAMCancelMonitor(%d)\n", fr->reqnum);
 
     /*
-     * destroy the request internally
+     * Verify the request
      */
     conn = fc->client;
-    ret = gamin_data_del_req(conn, fr->reqnum);
+    gamin_data_lock(conn);
+    ret = gamin_data_cancel(conn, fr->reqnum);
     if (ret < 0) {
         FAMErrno = FAM_ARG;
+	gamin_data_unlock(conn);
         return (-1);
     }
 
@@ -1331,6 +1401,7 @@ FAMCancelMonitor(FAMConnection * fc, const FAMRequest * fr)
      */
     ret = gamin_send_request(GAM_REQ_CANCEL, fc->fd, NULL,
                              (FAMRequest *) fr, NULL, fc->client, 0);
+    gamin_data_unlock(conn);
 
     if (ret != 0) {
         FAMErrno = FAM_CONNECT;
@@ -1393,7 +1464,9 @@ int FAMNoExists(FAMConnection *fc) {
     }
     conn = fc->client;
 
+    gamin_data_lock(conn);
     ret = gamin_data_no_exists(conn);
+    gamin_data_unlock(conn);
     if (ret < 0) {
 	GAM_DEBUG(DEBUG_INFO, "FAMNoExists() arg error\n");
         FAMErrno = FAM_ARG;
@@ -1440,8 +1513,11 @@ FAMDebug(FAMConnection *fc, const char *filename, FAMRequest * fr,
     /*
      * send debug message to the server
      */
+    gamin_data_lock(fc->client);
     ret = gamin_send_request(GAM_REQ_DEBUG, fc->fd, filename,
 			     fr, userData, fc->client, 0);
+    gamin_data_unlock(fc->client);
+
     if (debug_reqno == -1) {
         debug_reqno = fr->reqnum;
 	debug_userData = userData;
