@@ -36,6 +36,7 @@ int		gam_connection_close	(GamConnDataPtr conn);
 
 int		gam_connection_get_fd	(GamConnDataPtr conn);
 int		gam_connection_get_pid  (GamConnDataPtr conn);
+gchar *		gam_connection_get_pidname (GamConnDataPtr conn);
 GamConnState	gam_connection_get_state(GamConnDataPtr conn);
 int		gam_connection_get_data	(GamConnDataPtr conn,
 					 char **data,
@@ -47,6 +48,15 @@ int		gam_send_event		(GamConnDataPtr conn,
 					 int event,
 					 const char *path,
 					 int len);
+void		gam_queue_event		(GamConnDataPtr conn,
+					 int reqno,
+					 int event,
+					 const char *path,
+					 int len);
+int		gam_send_ack		(GamConnDataPtr conn,
+					 int reqno,
+					 const char *path,
+					 int len);
 void		gam_connections_debug	(void);
 #ifdef __cplusplus
 }
